@@ -631,6 +631,18 @@ class Array(metaclass=MetaArray):
         else:
             raise NotImplementedError
 
+    def to_nparray(self):
+        shape = self._shape
+        cshape = [shape[ii] for ii in self._order]
+        if hasattr(self._itemtype, "_dtype"):
+            arr = self._buffer.to_nparray(
+                self._offset + self._data_offset, self._itemtype._dtype, cshape
+            ).transpose(self._order)
+            assert arr.strides == self._strides
+            return arr
+        else:
+            raise NotImplementedError
+
     @classmethod
     def _gen_data_paths(cls, base=None):
         paths = []
@@ -671,6 +683,18 @@ class Array(metaclass=MetaArray):
     def _get_inner_types(cls):
         return [cls._itemtype]
 
+    def _to_json(self):
+        out = []
+        for v in self:  # TODO does not support multidimensional arrays
+            if hasattr(v, "_to_json"):
+                vdata = v._to_json()
+            else:
+                vdata = v
+            if self._has_refs and v is not None:
+                vdata = (v.__class__.__name__, vdata)
+            out.append(vdata)
+        return out
+
 
 def is_index(atype):
     return isinstance(atype, Index)
