@@ -1,26 +1,26 @@
 # copyright ################################# #
 # This file is part of the Xobjects Package.  #
-# Copyright (c) CERN, 2021.                   #
+# Copyright (c) CERN, 2022.                   #
 # ########################################### #
 
-import os
 import logging
 
 import numpy as np
+from typing import List
 
 from .context import (
+    ModuleNotAvailable,
+    SourceType,
     XBuffer,
     XContext,
-    ModuleNotAvailable,
+    _concatenate_sources,
     available,
     classes_from_kernels,
     sort_classes,
     sources_from_classes,
-    _concatenate_sources,
 )
-
-from .specialize_source import specialize_source
 from .linkedarray import BaseLinkedArray
+from .specialize_source import specialize_source
 
 log = logging.getLogger(__name__)
 
@@ -35,7 +35,7 @@ except ImportError:
     )
     cl = ModuleNotAvailable(
         message=(
-            "pyopencl is not installed. " "ContextPyopencl is not available!"
+            "pyopencl is not installed. ContextPyopencl is not available!"
         )
     )
     cl.Buffer = cl
@@ -44,8 +44,7 @@ except ImportError:
 
 from ._patch_pyopencl_array import _patch_pyopencl_array
 
-
-openclheader = [
+openclheader: List[SourceType] = [
     """\
 #ifndef XOBJ_STDINT
 typedef long int64_t;
@@ -170,87 +169,27 @@ class ContextPyopencl(XContext):
             )
         return i
 
-    def add_kernels(
+    def build_kernels(
         self,
-        sources=[],
-        kernels=[],
+        sources,
+        kernel_descriptions,
         specialize=True,
         apply_to_source=(),
         save_source_as=None,
         extra_cdef=None,
-        extra_classes=[],
-        extra_headers=[],
-        compile=True,
+        extra_classes=(),
+        extra_headers=(),
+        compile=True,  # noqa
     ):
-
-        """
-        Adds user-defined kernels to to the context. The kernel source
-        code is provided as a string and/or in source files and must contain
-        the kernel names defined in the kernel descriptions.
-        Args:
-            sources (list): List of source codes that are concatenated before
-                compilation. The list can contain strings (raw source code),
-                File objects and Path objects.
-            kernels (dict): Dictionary with the kernel descriptions
-                in the form given by the following examples. The descriptions
-                define the kernel names, the type and name of the arguments
-                and identify one input argument that defines the number of
-                threads to be launched (only on cuda/opencl).
-            specialize_code (bool): If True, the code is specialized using
-                annotations in the source code. Default is ``True``
-            save_source_as (str): Filename for saving the specialized source
-                code. Default is ```None```.
-        Example:
-
-        .. code-block:: python
-
-            # A simple kernel
-            src_code = '''
-            /*gpukern*/
-            void my_mul(const int n,
-                /*gpuglmem*/ const double* x1,
-                /*gpuglmem*/ const double* x2,
-                /*gpuglmem*/       double* y) {
-                int tid = 0 //vectorize_over tid
-                y[tid] = x1[tid] * x2[tid];
-                //end_vectorize
-                }
-            '''
-
-            # Prepare description
-            kernel_descriptions = {
-                "my_mul": xo.Kernel(
-                    args=[
-                        xo.Arg(xo.Int32, name="n"),
-                        xo.Arg(xo.Float64, pointer=True, const=True, name="x1"),
-                        xo.Arg(xo.Float64, pointer=True, const=True, name="x2"),
-                        xo.Arg(xo.Float64, pointer=True, const=False, name="y"),
-                    ],
-                    n_threads="n",
-                    ),
-            }
-
-            # Import kernel in context
-            ctx.add_kernels(
-                sources=[src_code],
-                kernels=kernel_descriptions,
-                save_source_as=None,
-            )
-
-            # With a1, a2, b being arrays on the context, the kernel
-            # can be called as follows:
-            ctx.kernels.my_mul(n=len(a1), x1=a1, x2=a2, y=b)
-        """
-
         if not compile:
             raise NotImplementedError("compile=False available only on CPU.")
 
-        classes = classes_from_kernels(kernels)
+        classes = classes_from_kernels(kernel_descriptions)
         classes.update(extra_classes)
         classes = sort_classes(classes)
         cls_sources = sources_from_classes(classes)
 
-        headers = openclheader + extra_headers
+        headers = openclheader + list(extra_headers)
 
         sources = headers + cls_sources + sources
 
@@ -270,18 +209,21 @@ class ContextPyopencl(XContext):
 
         prg = cl.Program(self.context, specialized_source).build()
 
-        for pyname, kernel in kernels.items():
+        out_kernels = {}
+        for pyname, kernel in kernel_descriptions.items():
             if kernel.c_name is None:
                 kernel.c_name = pyname
 
-            self.kernels[pyname] = KernelPyopencl(
+            out_kernels[pyname] = KernelPyopencl(
                 function=getattr(prg, kernel.c_name),
                 description=kernel,
                 context=self,
             )
 
-            self.kernels[pyname].source = source
-            self.kernels[pyname].specialized_source = specialized_source
+            out_kernels[pyname].source = source
+            out_kernels[pyname].specialized_source = specialized_source
+
+        return out_kernels
 
     def nparray_to_context_array(self, arr):
 
