@@ -233,11 +233,9 @@ class MetaStruct(type):
                         info._offsets = offsets
                         if len(extra) > 0:
                             info.extra = extra
-                    elif isinstance(arg, cls):
+                    elif isinstance(arg, cls):  # is the same size
                         info.size = arg._get_size()
-                        info._offsets = {
-                            kk: vv for kk, vv in arg._offsets.items()
-                        }
+                        info._offsets = arg._offsets
                     else:
                         raise ValueError(f"{arg} Not valid type for {cls}")
                 else:  # python argument
@@ -370,6 +368,15 @@ class Struct(metaclass=MetaStruct):
     def _to_dict(self):
         return {field.name: field.__get__(self) for field in self._fields}
 
+    def _to_json(self):
+        out = {}
+        for field in self._fields:
+            v = field.__get__(self)
+            if hasattr(v, "_to_json"):
+                v = v._to_json()
+            out[field.name] = v
+        return out
+
     def __iter__(self):
         for field in self._fields:
             yield field.name
